#  hammer_build_systems.py
#  Class containing all the methods to create VLSI flow build system infrastructure
#
#  See LICENSE for licence details.

from .driver import HammerDriver

import os
import sys
import textwrap
from typing import List, Dict, Tuple, Callable

def build_noop(driver: HammerDriver, append_error_func: Callable[[str], None]) -> dict:
    """
    Do nothing, just return the dependency graph.

    :param driver: The HammerDriver
    :return: The dependency graph
    """
    dependency_graph = driver.get_hierarchical_dependency_graph()
    return dependency_graph


def build_makefile(driver: HammerDriver, append_error_func: Callable[[str], None]) -> dict:
    """
    Build a Makefile include in the obj_dir called hammer.d. This is intended to be dynamically
    created and included into a top-level Makefile.

    The Makefile will contain targets for the following hammer actions, as well as any necessary
    bridge actions (xyz-to-abc):
        - pcb
        - syn
        - par
        - drc
        - lvs
        - sim-rtl
        - sim-syn
        - sim-par
        - power-rtl
        - power-syn
        - power-par
        - formal-syn
        - formal-par
        - timing-syn
        - timing-par

    For hierarchical flows, the syn, par, drc, lvs, sim, power, formal, and timing actions will all be suffixed with the name
    of the hierarchical modules (e.g. syn-Top, syn-SubModA, par-SubModA, etc.). The appropriate
    dependencies and bridge actions are automatically generated from the hierarchy provided in the
    Hammer IR.

    For actions that can be run at multiple points in the flow such as sim, the name of the target
    will include the action it is being run after (e.g. sim-syn, sim-par, etc.). With no suffix
    an rtl level simulation will be run.

    Additionally, "redo" steps are created (e.g. redo-par for flat designs or redo-par-Top for
    hierarchical), which allow the user to bypass the normal Makefile dependencies and force a
    rerun of a particular task. This is useful when the user wants to change an input Hammer IR file
    knowing it will not affect intermediate steps in the design.

    An example use case for integrating this file into a top flow is provided below. Be sure to use
    real tabs if copying this snippet!

    ```
    TOP ?= MyTop
    OBJ_DIR ?= $(abspath build-$(TOP))
    INPUT_CONFS ?= foo.yaml bar.yaml baz.yaml

    HAMMER_EXEC ?= ./mychip-vlsi.py

    .PHONY: all
    all: drc-$(TOP) lvs-$(TOP)

    GENERATED_CONF = $(OBJ_DIR)/input.yaml

    $(GENERATED_CONF):
        echo "synthesis.inputs.top_module: $(TOP)" > $@
        echo "pcb.inputs.top_module: $(TOP)" >> $@

    $(OBJ_DIR)/hammer.d: $(GENERATED_CONF)
        $(HAMMER_EXEC) -e env.yaml $(foreach x,$(INPUT_CONFS) $(GENERATED_CONF), -p $(x)) --obj_dir $(OBJ_DIR) build

    include $(OBJ_DIR)/hammer.d
    ```

    The generated Makefile has a few variables that are set if absent. This allows the user to override them without
    modifying hammer.d. They are listed as follows:
        - HAMMER_EXEC: This sets the actual python executable containing the HammerDriver main() function. It is set to
          the executable used to generate the Makefile by default.
        - HAMMER_DEPENDENCIES: The list of dependences to use for the initial syn and pcb targets. It is set to the set
          of all input configurations, environment settings, and input files by default.
        - HAMMER_*_DEPENDENCIES: There is a version of this variable for each action. It allows the user to have more
          fine grained depencies compared to the blunt HAMMER_DEPENDENCIES. It is set as a dependency for the respective
          action. Often used with clearing the global HAMMER_DEPENDENCIES.
        - HAMMER_EXTRA_ARGS: This is passed to the Hammer executable for all targets. This is unset by default.
          Its primary uses are for adding additional configuration files with -p, --to_step/until_step, and/or --from_step/
          after_step options. An example use is "make redo-par-Top HAMMER_EXTRA_ARGS="-p patch.yaml --from_step placement".

    :param driver: The HammerDriver
    :return: The dependency graph
    """
    dependency_graph = driver.get_hierarchical_dependency_graph()
    makefile = os.path.join(driver.obj_dir, "hammer.d")
    default_dependencies = driver.options.project_configs + driver.options.environment_configs
    default_dependencies.extend(list(driver.database.get_setting("synthesis.inputs.input_files", [])))
    # Resolve the canonical path for each dependency
    default_dependencies = [os.path.realpath(x) for x in default_dependencies]
    output = "HAMMER_EXEC ?= {}\n".format(os.path.realpath(sys.argv[0]))
    output += "HAMMER_DEPENDENCIES ?= {}\n\n".format(" ".join(default_dependencies))
    syn_deps = "$(HAMMER_DEPENDENCIES)"
    # Get the confs passed into this execution
    env_confs = " ".join(["-e " + os.path.realpath(x) for x in driver.options.environment_configs])
    proj_confs = " ".join(["-p " + os.path.realpath(x) for x in driver.options.project_configs])
    obj_dir = os.path.realpath(driver.obj_dir)

    # Global steps that are the same for hier or flat
    pcb_run_dir = os.path.join(obj_dir, "pcb-rundir")
    pcb_out = os.path.join(pcb_run_dir, "pcb-output-full.json")
    output += textwrap.dedent("""
        ####################################################################################
        ## Global steps
        ####################################################################################
        .PHONY: pcb
        pcb: {pcb_out}

        {pcb_out}: {syn_deps}
        \t$(HAMMER_EXEC) {env_confs} {all_inputs} --obj_dir {obj_dir} pcb

        """.format(pcb_out=pcb_out, syn_deps=syn_deps, env_confs=env_confs, all_inputs=proj_confs, obj_dir=obj_dir))

    make_text = textwrap.dedent("""
        ####################################################################################
        ## Steps for {mod}
        ####################################################################################
        .PHONY: sim-rtl{suffix} syn{suffix} syn-to-sim{suffix} sim-syn{suffix} syn-to-par{suffix} par{suffix} par-to-sim{suffix} sim-par{suffix} sim-par-to-power{suffix} par-to-power{suffix} power-par{suffix} power-rtl{suffix} sim-rtl-to-power{suffix} sim-syn-to-power{suffix} syn-to-power{suffix} power-syn{suffix} par-to-drc{suffix} drc{suffix} par-to-lvs{suffix} lvs{suffix} syn-to-formal{suffix} formal-syn{suffix} par-to-formal{suffix} formal-par{suffix} syn-to-timing{suffix} timing-syn{suffix} par-to-timing{suffix} timing-par{suffix}

        sim-rtl{suffix}          : {sim_rtl_out}
        syn{suffix}              : {syn_out}

        syn-to-sim{suffix}       : {sim_syn_in}
        sim-syn{suffix}          : {sim_syn_out}

        syn-to-par{suffix}       : {par_in}
        par{suffix}              : {par_out}

        par-to-sim{suffix}       : {sim_par_in}
        sim-par{suffix}          : {sim_par_out}

        sim-par-to-power{suffix} : {power_sim_par_in}
        par-to-power{suffix}     : {power_par_in}
        power-par{suffix}        : {power_par_out}

        sim-rtl-to-power{suffix} : {power_sim_rtl_in}
        power-rtl{suffix}        : {power_rtl_out}

        sim-syn-to-power{suffix} : {power_sim_syn_in}
        syn-to-power{suffix}     : {power_syn_in}
        power-syn{suffix}        : {power_syn_out}

        par-to-drc{suffix}       : {drc_in}
        drc{suffix}              : {drc_out}

        par-to-lvs{suffix}       : {lvs_in}
        lvs{suffix}              : {lvs_out}

        syn-to-formal{suffix}    : {formal_syn_in}
        formal-syn{suffix}       : {formal_syn_out}

        par-to-formal{suffix}    : {formal_par_in}
        formal-par{suffix}       : {formal_par_out}

        syn-to-timing{suffix}    : {timing_syn_in}
        timing-syn{suffix}       : {timing_syn_out}

        par-to-timing{suffix}    : {timing_par_in}
        timing-par{suffix}       : {timing_par_out}

        {par_to_syn}

        {sim_rtl_out}: {syn_deps} $(HAMMER_SIM_RTL_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} {p_sim_rtl_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_rtl_run_dir} --obj_dir {obj_dir} sim{suffix}

        {power_sim_rtl_in}: {sim_rtl_out}
        \t$(HAMMER_EXEC) {env_confs} -p {sim_rtl_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_rtl_in} --obj_dir {obj_dir} sim-to-power

        {power_rtl_out}: {power_sim_rtl_in} $(HAMMER_POWER_RTL_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_rtl_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_rtl_run_dir} --obj_dir {obj_dir} power{suffix}

        {syn_out}: {syn_deps} $(HAMMER_SYN_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} {p_syn_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} syn{suffix}

        {sim_syn_in}: {syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {sim_syn_in} --obj_dir {obj_dir} syn-to-sim

        {sim_syn_out}: {sim_syn_in} $(HAMMER_SIM_SYN_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {sim_syn_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_syn_run_dir} --obj_dir {obj_dir} sim{suffix}

        {power_sim_syn_in}: {sim_syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {sim_syn_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_syn_in} --obj_dir {obj_dir} sim-to-power

        {power_syn_in}: {syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {power_syn_in} --obj_dir {obj_dir} syn-to-power

        {power_syn_out}: {power_sim_syn_in} {power_syn_in} $(HAMMER_POWER_SYN_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_syn_in} -p {power_syn_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_syn_run_dir} --obj_dir {obj_dir} power{suffix}

        {par_in}: {syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {par_in} --obj_dir {obj_dir} syn-to-par

        {par_out}: {par_in} $(HAMMER_PAR_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {par_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} par{suffix}

        {sim_par_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {sim_par_in} --obj_dir {obj_dir} par-to-sim

        {sim_par_out}: {sim_par_in} $(HAMMER_SIM_PAR_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {sim_par_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_par_run_dir} --obj_dir {obj_dir} sim{suffix}

        {power_sim_par_in}: {sim_par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {sim_par_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_par_in} --obj_dir {obj_dir} sim-to-power

        {power_par_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {power_par_in} --obj_dir {obj_dir} par-to-power

        {power_par_out}: {power_sim_par_in} {power_par_in} $(HAMMER_POWER_PAR_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_par_in} -p {power_par_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_par_run_dir} --obj_dir {obj_dir} power{suffix}

        {drc_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {drc_in} --obj_dir {obj_dir} par-to-drc

        {drc_out}: {drc_in} $(HAMMER_DRC_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {drc_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} drc{suffix}

        {lvs_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {lvs_in} --obj_dir {obj_dir} par-to-lvs

        {lvs_out}: {lvs_in} $(HAMMER_LVS_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {lvs_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} lvs{suffix}

        {formal_syn_in}: {syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {formal_syn_in} --obj_dir {obj_dir} syn-to-formal

        {formal_syn_out}: {formal_syn_in} $(HAMMER_FORMAL_SYN_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {formal_syn_in} $(HAMMER_EXTRA_ARGS) --formal_rundir {formal_syn_run_dir} --obj_dir {obj_dir} formal{suffix}

        {formal_par_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {formal_par_in} --obj_dir {obj_dir} par-to-formal

        {formal_par_out}: {formal_syn_in} $(HAMMER_FORMAL_PAR_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {formal_par_in} $(HAMMER_EXTRA_ARGS) --formal_rundir {formal_par_run_dir} --obj_dir {obj_dir} formal{suffix}

        {timing_syn_in}: {syn_out}
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {timing_syn_in} --obj_dir {obj_dir} syn-to-timing

        {timing_syn_out}: {timing_syn_in} $(HAMMER_TIMING_SYN_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {timing_syn_in} $(HAMMER_EXTRA_ARGS) --timing_rundir {timing_syn_run_dir} --obj_dir {obj_dir} timing{suffix}

        {timing_par_in}: {par_out}
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {timing_par_in} --obj_dir {obj_dir} par-to-timing

        {timing_par_out}: {timing_syn_in} $(HAMMER_TIMING_PAR_DEPENDENCIES)
        \t$(HAMMER_EXEC) {env_confs} -p {timing_par_in} $(HAMMER_EXTRA_ARGS) --timing_rundir {timing_par_run_dir} --obj_dir {obj_dir} timing{suffix}

        # Redo steps
        # These intentionally break the dependency graph, but allow the flexibility to rerun a step after changing a config.
        # Hammer doesn't know what settings impact synthesis only, e.g., so these are for power-users who "know better."
        # The HAMMER_EXTRA_ARGS variable allows patching in of new configurations with -p or using --to_step or --from_step, for example.
        .PHONY: redo-sim-rtl{suffix} redo-sim-rtl-to-power{suffix} redo-syn{suffix} redo-syn-to-sim{suffix} redo-syn-to-power{suffix} redo-sim-syn{suffix} redo-sim-syn-to-power{suffix} redo-syn-to-par{suffix} redo-par{suffix} redo-par-to-sim{suffix} redo-sim-par{suffix} redo-sim-par-to-power{suffix} redo-par-to-power{suffix} redo-power-par{suffix} redo-par-to-drc{suffix} redo-drc{suffix} redo-par-to-lvs{suffix} redo-lvs{suffix} redo-syn-to-formal{suffix} redo-formal-syn{suffix} redo-par-to-formal{suffix} redo-formal-par{suffix} redo-syn-to-timing{suffix} redo-timing-syn{suffix} redo-par-to-timing{suffix} redo-timing-par{suffix}

        redo-sim-rtl{suffix}:
        \t$(HAMMER_EXEC) {env_confs} {p_sim_rtl_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_rtl_run_dir} --obj_dir {obj_dir} sim{suffix}

        redo-sim-rtl-to-power{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {sim_rtl_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_rtl_in} --obj_dir {obj_dir} sim-to-power

        redo-power-rtl{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_rtl_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_rtl_run_dir} --obj_dir {obj_dir} power{suffix}

        redo-syn{suffix}:
        \t$(HAMMER_EXEC) {env_confs} {p_syn_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} syn{suffix}

        redo-syn-to-sim{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {sim_syn_in} --obj_dir {obj_dir} syn-to-sim

        redo-syn-to-power{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {power_syn_in} --obj_dir {obj_dir} syn-to-power

        redo-sim-syn{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {sim_syn_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_syn_run_dir} --obj_dir {obj_dir} sim{suffix}

        redo-sim-syn-to-power{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {sim_syn_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_syn_in} --obj_dir {obj_dir} sim-to-power

        redo-syn-to-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {par_in} --obj_dir {obj_dir} syn-to-par

        redo-power-syn{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_syn_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_syn_run_dir} --obj_dir {obj_dir} power{suffix}

        redo-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} par{suffix}

        redo-par-to-sim{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {sim_par_in} --obj_dir {obj_dir} par-to-sim

        redo-sim-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {sim_par_in} $(HAMMER_EXTRA_ARGS) --sim_rundir {sim_par_run_dir} --obj_dir {obj_dir} sim{suffix}

        redo-sim-par-to-power{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {sim_par_out} $(HAMMER_EXTRA_ARGS) -o {power_sim_par_in} --obj_dir {obj_dir} sim-to-power

        redo-par-to-power{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {power_par_in} --obj_dir {obj_dir} par-to-power

        redo-power-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {power_sim_par_in} -p {power_par_in} $(HAMMER_EXTRA_ARGS) --power_rundir {power_par_run_dir} --obj_dir {obj_dir} power{suffix}

        redo-par-to-drc{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {drc_in} --obj_dir {obj_dir} par-to-drc

        redo-drc{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {drc_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} drc{suffix}

        redo-par-to-lvs{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {lvs_in} --obj_dir {obj_dir} par-to-lvs

        redo-lvs{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {lvs_in} $(HAMMER_EXTRA_ARGS) --obj_dir {obj_dir} lvs{suffix}

        redo-syn-to-formal{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {formal_syn_in} --obj_dir {obj_dir} syn-to-formal

        redo-formal-syn{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {formal_syn_in} $(HAMMER_EXTRA_ARGS) --formal_rundir {formal_syn_run_dir} --obj_dir {obj_dir} formal{suffix}

        redo-par-to-formal{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {formal_par_in} --obj_dir {obj_dir} par-to-formal

        redo-formal-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {formal_par_in} $(HAMMER_EXTRA_ARGS) --formal_rundir {formal_par_run_dir} --obj_dir {obj_dir} formal{suffix}

        redo-syn-to-timing{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {syn_out} $(HAMMER_EXTRA_ARGS) -o {timing_syn_in} --obj_dir {obj_dir} syn-to-timing

        redo-timing-syn{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {timing_syn_in} $(HAMMER_EXTRA_ARGS) --timing_rundir {timing_syn_run_dir} --obj_dir {obj_dir} timing{suffix}

        redo-par-to-timing{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {par_out} $(HAMMER_EXTRA_ARGS) -o {timing_par_in} --obj_dir {obj_dir} par-to-timing

        redo-timing-par{suffix}:
        \t$(HAMMER_EXEC) {env_confs} -p {timing_par_in} $(HAMMER_EXTRA_ARGS) --timing_rundir {timing_par_run_dir} --obj_dir {obj_dir} timing{suffix}

        """)

    if not dependency_graph:
        # Flat flow
        top_module = str(driver.database.get_setting("synthesis.inputs.top_module"))

        # TODO make this DRY
        sim_rtl_run_dir = os.path.join(obj_dir, "sim-rtl-rundir")
        power_rtl_run_dir = os.path.join(obj_dir, "power-rtl-rundir")
        syn_run_dir = os.path.join(obj_dir, "syn-rundir")
        sim_syn_run_dir = os.path.join(obj_dir, "sim-syn-rundir")
        power_syn_run_dir = os.path.join(obj_dir, "power-syn-rundir")
        par_run_dir = os.path.join(obj_dir, "par-rundir")
        sim_par_run_dir = os.path.join(obj_dir, "sim-par-rundir")
        power_par_run_dir = os.path.join(obj_dir, "power-par-rundir")
        drc_run_dir = os.path.join(obj_dir, "drc-rundir")
        lvs_run_dir = os.path.join(obj_dir, "lvs-rundir")
        formal_syn_run_dir = os.path.join(obj_dir, "formal-syn-rundir")
        formal_par_run_dir = os.path.join(obj_dir, "formal-par-rundir")
        timing_syn_run_dir = os.path.join(obj_dir, "timing-syn-rundir")
        timing_par_run_dir = os.path.join(obj_dir, "timing-par-rundir")

        p_sim_rtl_in = proj_confs
        sim_rtl_out = os.path.join(sim_rtl_run_dir, "sim-output-full.json")
        power_sim_rtl_in = os.path.join(obj_dir, "power-sim-rtl-input.json")
        #power_rtl_in = os.path.join(obj_dir, "power-rtl-input.json")
        power_rtl_out = os.path.join(power_rtl_run_dir, "power-output-full.json")
        p_syn_in = proj_confs
        syn_out = os.path.join(syn_run_dir, "syn-output-full.json")
        sim_syn_in = os.path.join(obj_dir, "sim-syn-input.json")
        sim_syn_out = os.path.join(sim_syn_run_dir, "sim-output-full.json")
        power_sim_syn_in = os.path.join(obj_dir, "power-sim-syn-input.json")
        power_syn_in = os.path.join(obj_dir, "power-syn-input.json")
        power_syn_out = os.path.join(power_syn_run_dir, "power-output-full.json")
        par_in = os.path.join(obj_dir, "par-input.json")
        par_out = os.path.join(par_run_dir, "par-output-full.json")
        sim_par_in = os.path.join(obj_dir, "sim-par-input.json")
        sim_par_out = os.path.join(sim_par_run_dir, "sim-output-full.json")
        power_sim_par_in = os.path.join(obj_dir, "power-sim-par-input.json")
        power_par_in = os.path.join(obj_dir, "power-par-input.json")
        power_par_out = os.path.join(power_par_run_dir, "power-output-full.json")
        drc_in = os.path.join(obj_dir, "drc-input.json")
        drc_out = os.path.join(drc_run_dir, "drc-output-full.json")
        lvs_in = os.path.join(obj_dir, "lvs-input.json")
        lvs_out = os.path.join(lvs_run_dir, "lvs-output-full.json")
        formal_syn_in = os.path.join(obj_dir, "formal-syn-input.json")
        formal_syn_out = os.path.join(formal_syn_run_dir, "formal-output-full.json")
        formal_par_in = os.path.join(obj_dir, "formal-par-input.json")
        formal_par_out = os.path.join(formal_par_run_dir, "formal-output-full.json")
        timing_syn_in = os.path.join(obj_dir, "timing-syn-input.json")
        timing_syn_out = os.path.join(timing_syn_run_dir, "timing-output-full.json")
        timing_par_in = os.path.join(obj_dir, "timing-par-input.json")
        timing_par_out = os.path.join(timing_par_run_dir, "timing-output-full.json")

        par_to_syn = ""

        output += make_text.format(suffix="", mod=top_module, env_confs=env_confs, obj_dir=obj_dir, syn_deps=syn_deps,
            par_to_syn=par_to_syn,
            p_sim_rtl_in=p_sim_rtl_in, sim_rtl_out=sim_rtl_out, sim_rtl_run_dir=sim_rtl_run_dir,
            sim_syn_in=sim_syn_in, sim_syn_out=sim_syn_out, sim_syn_run_dir=sim_syn_run_dir,
            power_sim_rtl_in=power_sim_rtl_in, power_rtl_out=power_rtl_out, power_rtl_run_dir=power_rtl_run_dir,
            sim_par_in=sim_par_in, sim_par_out=sim_par_out, sim_par_run_dir=sim_par_run_dir,
            p_syn_in=p_syn_in, syn_out=syn_out, par_in=par_in, par_out=par_out,
            power_sim_syn_in=power_sim_syn_in, power_syn_in=power_syn_in, power_syn_out=power_syn_out, power_syn_run_dir=power_syn_run_dir,
            power_sim_par_in=power_sim_par_in, power_par_in=power_par_in, power_par_out=power_par_out, power_par_run_dir=power_par_run_dir,
            drc_in=drc_in, drc_out=drc_out, lvs_in=lvs_in, lvs_out=lvs_out,
            formal_syn_in=formal_syn_in, formal_syn_out=formal_syn_out, formal_syn_run_dir=formal_syn_run_dir,
            formal_par_in=formal_par_in, formal_par_out=formal_par_out, formal_par_run_dir=formal_par_run_dir,
            timing_syn_in=timing_syn_in, timing_syn_out=timing_syn_out, timing_syn_run_dir=timing_syn_run_dir,
            timing_par_in=timing_par_in, timing_par_out=timing_par_out, timing_par_run_dir=timing_par_run_dir)
    else:
        # Hierarchical flow
        for node, edges in dependency_graph.items():
            out_edges = edges[1]

            # TODO make this DRY
            sim_rtl_run_dir = os.path.join(obj_dir, "sim-rtl-" + node)
            power_rtl_run_dir = os.path.join(obj_dir, "power-rtl-" + node)
            syn_run_dir = os.path.join(obj_dir, "syn-" + node)
            sim_syn_run_dir = os.path.join(obj_dir, "sim-syn-" + node)
            power_syn_run_dir = os.path.join(obj_dir, "power-syn-" + node)
            par_run_dir = os.path.join(obj_dir, "par-" + node)
            sim_par_run_dir = os.path.join(obj_dir, "sim-par-" + node)
            power_par_run_dir = os.path.join(obj_dir, "power-par-" + node)
            drc_run_dir = os.path.join(obj_dir, "drc-" + node)
            lvs_run_dir = os.path.join(obj_dir, "lvs-" + node)
            formal_syn_run_dir = os.path.join(obj_dir, "formal-syn-" + node)
            formal_par_run_dir = os.path.join(obj_dir, "formal-par-" + node)
            timing_syn_run_dir = os.path.join(obj_dir, "timing-syn-" + node)
            timing_par_run_dir = os.path.join(obj_dir, "timing-par-" + node)

            p_sim_rtl_in = proj_confs
            sim_rtl_out = os.path.join(sim_rtl_run_dir, "sim-output-full.json")
            power_sim_rtl_in = os.path.join(obj_dir, "power-sim-rtl-{}-input.json".format(node))
            #power_rtl_in = os.path.join(obj_dir, "power-rtl-{}-input.json".format(node))
            power_rtl_out = os.path.join(power_rtl_run_dir, "power-output-full.json")
            p_syn_in = proj_confs
            syn_out = os.path.join(syn_run_dir, "syn-output-full.json")
            sim_syn_in = os.path.join(obj_dir, "sim-syn-{}-input.json".format(node))
            sim_syn_out = os.path.join(sim_syn_run_dir, "sim-output-full.json")
            power_sim_syn_in = os.path.join(obj_dir, "power-sim-syn-{}-input.json".format(node))
            power_syn_in = os.path.join(obj_dir, "power-syn-{}-input.json".format(node))
            power_syn_out = os.path.join(power_syn_run_dir, "power-output-full.json")
            par_in = os.path.join(obj_dir, "par-{}-input.json".format(node))
            par_out = os.path.join(par_run_dir, "par-output-full.json")
            sim_par_in = os.path.join(obj_dir, "sim-par-{}-input.json".format(node))
            sim_par_out = os.path.join(sim_par_run_dir, "sim-output-full.json")
            power_sim_par_in = os.path.join(obj_dir, "power-sim-par-{}-input.json".format(node))
            power_par_in = os.path.join(obj_dir, "power-par-{}-input.json".format(node))
            power_par_out = os.path.join(power_par_run_dir, "power-output-full.json")
            drc_in = os.path.join(obj_dir, "drc-{}-input.json".format(node))
            drc_out = os.path.join(drc_run_dir, "drc-output-full.json")
            lvs_in = os.path.join(obj_dir, "lvs-{}-input.json".format(node))
            lvs_out = os.path.join(lvs_run_dir, "lvs-output-full.json")
            formal_syn_in = os.path.join(obj_dir, "formal-syn-{}-input.json".format(node))
            formal_syn_out = os.path.join(formal_syn_run_dir, "formal-output-full.json")
            formal_par_in = os.path.join(obj_dir, "formal-par-{}-input.json".format(node))
            formal_par_out = os.path.join(formal_par_run_dir, "formal-output-full.json")
            timing_syn_in = os.path.join(obj_dir, "timing-syn-{}-input.json".format(node))
            timing_syn_out = os.path.join(timing_syn_run_dir, "timing-output-full.json")
            timing_par_in = os.path.join(obj_dir, "timing-par-{}-input.json".format(node))
            timing_par_out = os.path.join(timing_par_run_dir, "timing-output-full.json")

            # need to revert this each time
            syn_deps = "$(HAMMER_DEPENDENCIES)"
            par_to_syn = ""
            if len(out_edges) > 0:
                syn_deps = os.path.join(obj_dir, "syn-{}-input.json".format(node))
                p_syn_in = "-p {}".format(syn_deps)
                out_confs = [os.path.join(obj_dir, "par-" + x, "par-output-full.json") for x in out_edges]
                prereqs = " ".join(out_confs)
                pstring = " ".join(["-p " + x for x in out_confs])
                par_to_syn = textwrap.dedent("""
                    {syn_deps}: {prereqs}
                    \t$(HAMMER_EXEC) {env_confs} {pstring} -o {syn_deps} --obj_dir {obj_dir} hier-par-to-syn
                    """.format(syn_deps=syn_deps, prereqs=prereqs, env_confs=env_confs, pstring=pstring,
                    obj_dir=obj_dir))

            output += make_text.format(suffix="-"+node, mod=node, env_confs=env_confs, obj_dir=obj_dir, syn_deps=syn_deps,
                par_to_syn=par_to_syn,
                p_sim_rtl_in=p_sim_rtl_in, sim_rtl_out=sim_rtl_out, sim_rtl_run_dir=sim_rtl_run_dir,
                power_sim_rtl_in=power_sim_rtl_in, power_rtl_out=power_rtl_out, power_rtl_run_dir=power_rtl_run_dir,
                sim_syn_in=sim_syn_in, sim_syn_out=sim_syn_out, sim_syn_run_dir=sim_syn_run_dir,
                sim_par_in=sim_par_in, sim_par_out=sim_par_out, sim_par_run_dir=sim_par_run_dir,
                p_syn_in=p_syn_in, syn_out=syn_out, par_in=par_in, par_out=par_out,
                power_sim_syn_in=power_sim_syn_in, power_syn_in=power_syn_in, power_syn_out=power_syn_out, power_syn_run_dir=power_syn_run_dir,
                power_sim_par_in=power_sim_par_in, power_par_in=power_par_in, power_par_out=power_par_out, power_par_run_dir=power_par_run_dir,
                drc_in=drc_in, drc_out=drc_out, lvs_in=lvs_in, lvs_out=lvs_out,
                formal_syn_in=formal_syn_in, formal_syn_out=formal_syn_out, formal_syn_run_dir=formal_syn_run_dir,
                formal_par_in=formal_par_in, formal_par_out=formal_par_out, formal_par_run_dir=formal_par_run_dir,
                timing_syn_in=timing_syn_in, timing_syn_out=timing_syn_out, timing_syn_run_dir=timing_syn_run_dir,
                timing_par_in=timing_par_in, timing_par_out=timing_par_out, timing_par_run_dir=timing_par_run_dir)

    with open(makefile, "w") as f:
        f.write(output)

    return dependency_graph

BuildSystems = {
    "make": build_makefile,
    "none": build_noop
}  # type: Dict[str, Callable[[HammerDriver, Callable[[str], None]], dict]]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hammer',
 'hammer.common',
 'hammer.common.cadence',
 'hammer.common.mentor',
 'hammer.common.openroad',
 'hammer.common.synopsys',
 'hammer.common.vivado',
 'hammer.config',
 'hammer.drc.icv',
 'hammer.drc.magic',
 'hammer.drc.mockdrc',
 'hammer.drc.nop',
 'hammer.drc.openroad',
 'hammer.formal.conformal',
 'hammer.logging',
 'hammer.lvs.icv',
 'hammer.lvs.mocklvs',
 'hammer.lvs.netgen',
 'hammer.lvs.nop',
 'hammer.par.innovus',
 'hammer.par.mockpar',
 'hammer.par.nop',
 'hammer.par.openroad',
 'hammer.par.vivado',
 'hammer.pcb.generic',
 'hammer.power.joules',
 'hammer.power.voltus',
 'hammer.shell',
 'hammer.sim.mocksim',
 'hammer.sim.vcs',
 'hammer.sim.xcelium',
 'hammer.sram_generator.mocksram_generator',
 'hammer.sram_generator.nop',
 'hammer.synthesis.genus',
 'hammer.synthesis.mocksynth',
 'hammer.synthesis.nop',
 'hammer.synthesis.vivado',
 'hammer.synthesis.yosys',
 'hammer.tech',
 'hammer.technology.asap7',
 'hammer.technology.asap7.sram_compiler',
 'hammer.technology.nangate45',
 'hammer.technology.nangate45.sram_compiler',
 'hammer.technology.nop',
 'hammer.technology.sky130',
 'hammer.technology.sky130.extra',
 'hammer.technology.sky130.extra.openram',
 'hammer.technology.sky130.extra.sky130-tech-gen-files',
 'hammer.technology.sky130.extra.sram22',
 'hammer.technology.sky130.sram_compiler',
 'hammer.timing.tempus',
 'hammer.utils',
 'hammer.vlsi',
 'hammer.vlsi.vendor']

package_data = \
{'': ['*'],
 'hammer': ['technology/vivado/*'],
 'hammer.synthesis.vivado': ['file_templates/*'],
 'hammer.technology.asap7.sram_compiler': ['memories/gds/*',
                                           'memories/lef/*',
                                           'memories/lib/SRAM1RW1024x16_lib/*',
                                           'memories/lib/SRAM1RW1024x17_lib/*',
                                           'memories/lib/SRAM1RW1024x32_lib/*',
                                           'memories/lib/SRAM1RW1024x37_lib/*',
                                           'memories/lib/SRAM1RW1024x38_lib/*',
                                           'memories/lib/SRAM1RW1024x44_lib/*',
                                           'memories/lib/SRAM1RW1024x64_lib/*',
                                           'memories/lib/SRAM1RW1024x8_lib/*',
                                           'memories/lib/SRAM1RW128x12_lib/*',
                                           'memories/lib/SRAM1RW128x14_lib/*',
                                           'memories/lib/SRAM1RW128x22_lib/*',
                                           'memories/lib/SRAM1RW128x40_lib/*',
                                           'memories/lib/SRAM1RW128x46_lib/*',
                                           'memories/lib/SRAM1RW128x48_lib/*',
                                           'memories/lib/SRAM1RW128x64_lib/*',
                                           'memories/lib/SRAM1RW128x8_lib/*',
                                           'memories/lib/SRAM1RW2048x8_lib/*',
                                           'memories/lib/SRAM1RW256x128_lib/*',
                                           'memories/lib/SRAM1RW256x32_lib/*',
                                           'memories/lib/SRAM1RW256x46_lib/*',
                                           'memories/lib/SRAM1RW256x48_lib/*',
                                           'memories/lib/SRAM1RW256x64_lib/*',
                                           'memories/lib/SRAM1RW256x8_lib/*',
                                           'memories/lib/SRAM1RW32x50_lib/*',
                                           'memories/lib/SRAM1RW4096x16_lib/*',
                                           'memories/lib/SRAM1RW4096x8_lib/*',
                                           'memories/lib/SRAM1RW512x128_lib/*',
                                           'memories/lib/SRAM1RW512x32_lib/*',
                                           'memories/lib/SRAM1RW512x64_lib/*',
                                           'memories/lib/SRAM1RW512x8_lib/*',
                                           'memories/lib/SRAM1RW64x128_lib/*',
                                           'memories/lib/SRAM1RW64x20_lib/*',
                                           'memories/lib/SRAM1RW64x21_lib/*',
                                           'memories/lib/SRAM1RW64x22_lib/*',
                                           'memories/lib/SRAM1RW64x32_lib/*',
                                           'memories/lib/SRAM1RW64x34_lib/*',
                                           'memories/lib/SRAM1RW64x8_lib/*',
                                           'memories/lib/SRAM2RW128x16_lib/*',
                                           'memories/lib/SRAM2RW128x32_lib/*',
                                           'memories/lib/SRAM2RW128x4_lib/*',
                                           'memories/lib/SRAM2RW128x8_lib/*',
                                           'memories/lib/SRAM2RW16x16_lib/*',
                                           'memories/lib/SRAM2RW16x32_lib/*',
                                           'memories/lib/SRAM2RW16x4_lib/*',
                                           'memories/lib/SRAM2RW16x8_lib/*',
                                           'memories/lib/SRAM2RW32x16_lib/*',
                                           'memories/lib/SRAM2RW32x22_lib/*',
                                           'memories/lib/SRAM2RW32x32_lib/*',
                                           'memories/lib/SRAM2RW32x33_lib/*',
                                           'memories/lib/SRAM2RW32x39_lib/*',
                                           'memories/lib/SRAM2RW32x4_lib/*',
                                           'memories/lib/SRAM2RW32x8_lib/*',
                                           'memories/lib/SRAM2RW64x16_lib/*',
                                           'memories/lib/SRAM2RW64x24_lib/*',
                                           'memories/lib/SRAM2RW64x32_lib/*',
                                           'memories/lib/SRAM2RW64x4_lib/*',
                                           'memories/lib/SRAM2RW64x8_lib/*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'numpy>=1.23.0,<2.0.0',
 'pydantic>=1.9.2,<2.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0']

extras_require = \
{'asap7': ['gdspy==1.4'], 'asap7-gdstk': ['gdstk>=0.9.0,<0.10.0']}

entry_points = \
{'console_scripts': ['asap7_gds_scale = hammer.technology.asap7.gds_scale:main',
                     'get-config = hammer.shell.get_config:main',
                     'hammer-generate-mdf = '
                     'hammer.shell.hammer_generate_mdf:main',
                     'hammer-shell-test = hammer.shell.hammer_shell_test:main',
                     'hammer-vlsi = hammer.shell.hammer_vlsi:main',
                     'readlink-array = hammer.shell.readlink_array:main',
                     'yaml2json = hammer.shell.yaml2json:main']}

setup_kwargs = {
    'name': 'hammer-vlsi',
    'version': '1.1.1',
    'description': 'Hammer is a physical design framework that wraps around vendor specific technologies and tools to provide a single API to create ASICs.',
    'long_description': "![Logo design by @kenhoberkeley](https://github.com/ucb-bar/hammer/raw/master/doc/logo_transparent.png)\n\n![Hammer PR CI](https://github.com/ucb-bar/hammer/actions/workflows/pr.yml/badge.svg?event=push) ![Hammer publish CI](https://github.com/ucb-bar/hammer/actions/workflows/publish.yml/badge.svg)\n\nHighly Agile Masks Made Effortlessly from RTL (Hammer)\n=============================================\n\nThis is the master (unstable) development branch.\n\nDocumentation\n=============\nThe documentation is found at https://hammer-vlsi.readthedocs.io\nDoc sources are available at `doc/` for offline reading.\n\nHammer currently requires Python 3.9+.\n\nFor a deeper dive into available options and environment configuration:\n\n* [Core Hammer settings](hammer/config/defaults.yml)\n* [Setup documentation](https://hammer-vlsi.readthedocs.io/en/latest/Hammer-Basics/Hammer-Setup.html)\n* [Hammer technology library schema](https://hammer-vlsi.readthedocs.io/en/latest/Technology/Tech-json.html#full-schema)\n* For CAD tool settings, please see the relevant `defaults.yml` for those plugins.\n\nHammer is an integral component of UC Berkeley Architecture Research's [Chipyard framework](https://github.com/ucb-bar/chipyard).\nUseful documentation for how an example block is pushed through the VLSI flow with Hammer in the free ASAP7 and Sky130 PDKs is at https://chipyard.readthedocs.io/en/latest/VLSI.\n\nHistory\n=======\nThe list of contributors can be found in the Git history of the project, or online at https://github.com/ucb-bar/hammer/graphs/contributors\n\nThe Hammer project builds upon the legacy of the PLSI project by Palmer Dabbelt, a previous project which aimed to build a portable VLSI flow. The Hammer project is grateful for the feedback and lessons learned which provided valuable insight that ultimately led to the design and software architecture of Hammer as it is now.\n\nBelow is the list of Hammer publications in reverse chronological order:\n- [Hammer: a modular and reusable physical design flow tool (DAC 2022)](https://dl.acm.org/doi/abs/10.1145/3489517.3530672)\n- [A Methodology for Reusable Physical Design (ISQED 2020)](https://ieeexplore.ieee.org/document/9136999)\n- [PLSI: A Portable VLSI Flow (Master's Thesis, 2017)](https://www2.eecs.berkeley.edu/Pubs/TechRpts/2017/EECS-2017-77.html)\n",
    'author': 'Colin Schmidt',
    'author_email': 'colin.schmidt@sifive.com',
    'maintainer': 'Harrison Liew',
    'maintainer_email': 'harrisonliew@berkeley.edu',
    'url': 'https://github.com/ucb-bar/hammer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#  nop.py
#  No-op DRC tool.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerDRCTool, DummyHammerTool
from typing import List, Dict


class NopDRC(HammerDRCTool, DummyHammerTool):
    def globally_waived_drc_rules(self) -> List[str]:
        return []

    def drc_results_pre_waived(self) -> Dict[str, int]:
        return {}

    def fill_outputs(self) -> bool:
        return True


tool = NopDRC



#  hammer_vlsi.py
#  Main entry point to the hammer_vlsi library.
#
#  See LICENSE for licence details.

# Just import everything that the public hammer_vlsi module should see.

from . import units

from .hooks import *

from .hammer_vlsi_impl import *

from .hammer_tool import *

from .constraints import *

from .driver import *

from .cli_driver import CLIDriver

from .submit_command import *



#  submit_command.py
#  Hammer job submit command APIs.
#
#  See LICENSE for licence details.

# pylint: disable=bad-continuation

import atexit
import subprocess
import termios
import sys
import datetime
from abc import abstractmethod
from functools import reduce
from typing import Any, Dict, List, Tuple, NamedTuple, Optional

from hammer.config import HammerDatabase
from hammer.logging import HammerVLSILoggingContext
from hammer.utils import add_dicts, get_or_else

__all__ = ['HammerSubmitCommand', 'HammerLocalSubmitCommand',
           'HammerLSFSettings', 'HammerLSFSubmitCommand']


class HammerSubmitCommand:

    @abstractmethod
    def submit(self, args: List[str], env: Dict[str, str],
               logger: HammerVLSILoggingContext, cwd: Optional[str] = None) -> Tuple[str, int]:
        """
        Submit the job to the job submission system. This function MUST block
        until the command is complete.

        :param args: Command-line to run; each item in the list is one token.
                     The first token should be the command to run.
        :param env: The environment variables to set for the command
        :param logger: The logging context
        :param cwd: Working directory (leave as None to use the current working directory).
        :return: Tuple of the command output and a return code
        """
        pass

    @abstractmethod
    def read_settings(self, settings: Dict[str, Any], tool_namespace: str) -> None:
        """
        Read the settings object (a Dict[str, Any]) into meaningful class variables.

        :param settings: A Dict[str, Any] comprising the settings for this command.
        :param tool_namespace: The namespace for the tool (useful for logging).
        """
        pass

    @staticmethod
    def get(tool_namespace: str, database: HammerDatabase) -> "HammerSubmitCommand":
        """
        Get a concrete instance of a HammerSubmitCommand for a tool

        :param tool_namespace: The tool namespace to use when querying the
                               HammerDatabase (e.g. "synthesis" or "par").
        :param database: The HammerDatabase object with tool settings
        """

        submit_command_mode = database.get_setting(tool_namespace + ".submit.command",
                                                   nullvalue="none")
        # pylint: disable=line-too-long
        submit_command_settings = database.get_setting(tool_namespace + ".submit.settings",
                                                       nullvalue=[])  # type: List[Dict[str, Dict[str, Any]]]

        # Settings is a List[Dict[str, Dict[str, Any]]] object. The first Dict
        # key is the submit command name.
        # Its value is a Dict[str, Any] comprising the settings for that command.
        # The top-level list elements are merged from 0 to the last index, with
        # later indices overriding previous entries.
        def combine_settings(settings: List[Dict[str, Dict[str, Any]]], key: str) -> Dict[str, Any]:
            return reduce(add_dicts, map(lambda d: d[key], settings), {})

        submit_command = None  # type: Optional[HammerSubmitCommand]
        if submit_command_mode in {"none", "local"}:  # pylint: disable=no-else-return
            # Do not read the options, return immediately
            return HammerLocalSubmitCommand()
        elif submit_command_mode == "lsf":
            submit_command = HammerLSFSubmitCommand()
        else:
            raise NotImplementedError(
                "Submit command key for {0}: {1} is not implemented".format(
                    tool_namespace, submit_command_mode))

        submit_command.read_settings(
            combine_settings(submit_command_settings, submit_command_mode),
            tool_namespace)
        return submit_command

    @staticmethod
    def get_program_tag(args: List[str], program_name_length: int = 14,
                        arg_display_len: int = 16) -> str:
        """
        Get a short "tag" of the program for easier display in the log.
        :param args: Arguments for subprocess
        :param program_name_length: Capture last 14 (default) characters of
                                    the command name.
        :param arg_display_len: How many characters of args to display after prog_name
        :return: Short tag of the program.
        """
        if len(args[0]) <= program_name_length:
            prog_name = args[0]
        else:
            prog_name = "..." + args[0][len(args[0]) - program_name_length:]

        remaining_args = " ".join(args[1:])
        if len(remaining_args) < arg_display_len:
            prog_args = remaining_args
        else:
            prog_args = remaining_args[0:arg_display_len - 1] + "..."

        return prog_name + " " + prog_args


class HammerLocalSubmitCommand(HammerSubmitCommand):

    def submit(self, args: List[str], env: Dict[str, str],
               logger: HammerVLSILoggingContext, cwd: Optional[str] = None) -> Tuple[str, int]:
        # Just run the command on this host.

        prog_tag = self.get_program_tag(args)
        term_settings = termios.tcgetattr(sys.stdin.fileno()) if sys.stdin.isatty() else None

        logger.debug("Executing subprocess: " + ' '.join(args))
        subprocess_logger = logger.context("Exec " + prog_tag)
        proc = subprocess.Popen(args, shell=False, stderr=subprocess.STDOUT,
                                stdout=subprocess.PIPE, env=env, cwd=cwd)
        # These are run in reverse order of registration
        # Terminate is first to allow for the possibility of graceful shutdown
        # Then the program will be forceably terminated and we restore the terminal settings
        if term_settings is not None:
            atexit.register(termios.tcsetattr, sys.stdin.fileno(), termios.TCSANOW, term_settings)
        atexit.register(proc.kill)
        atexit.register(proc.terminate)

        output_buf = ""
        # Log output and also capture output at the same time.
        so = proc.stdout
        assert so is not None
        while True:
            line = so.readline().decode("utf-8")
            if line != '':
                subprocess_logger.debug(line.rstrip())
                output_buf += line
            else:
                break
        # check errors
        proc.communicate()

        return output_buf, proc.returncode

    def read_settings(self, settings: Dict[str, Any], tool_namespace: str) -> None:
        # Should never get here
        raise ValueError("Local submission command does not have settings")


class HammerLSFSettings(NamedTuple('HammerLSFSettings', [
    ('bsub_binary', str),
    ('num_cpus', Optional[int]),
    ('queue', Optional[str]),
    ('log_file', Optional[str]),
    ('extra_args', List[str])
])):
    __slots__ = ()

    @staticmethod
    def from_setting(settings: Dict[str, Any]) -> "HammerLSFSettings":
        if not isinstance(settings, dict):
            raise ValueError("Must be a dictionary")
        try:
            bsub_binary = settings["bsub_binary"]
        except KeyError:
            raise ValueError("Missing mandatory key bsub_binary for LSF settings.")
        try:
            num_cpus = settings["num_cpus"]
        except KeyError:
            num_cpus = None
        try:
            queue = settings["queue"]
        except KeyError:
            queue = None
        try:
            log_file = settings["log_file"]
        except KeyError:
            log_file = None

        return HammerLSFSettings(
            bsub_binary=bsub_binary,
            num_cpus=num_cpus,
            queue=queue,
            log_file=log_file,
            extra_args=get_or_else(settings["extra_args"], [])
        )


class HammerLSFSubmitCommand(HammerSubmitCommand):

    # TODO(johnwright): log the command output

    @property
    def settings(self) -> HammerLSFSettings:
        if not hasattr(self, "_settings"):
            raise ValueError("Nothing set for settings yet")
        return getattr(self, "_settings")

    @settings.setter
    def settings(self, value: HammerLSFSettings) -> None:
        """
        Set the settings class variable

        :param value: The HammerLSFSettings NapedTuple to use
        """
        setattr(self, "_settings", value)

    def read_settings(self, settings: Dict[str, Any], tool_namespace: str) -> None:  # pylint: disable=unused-argument
        self.settings = HammerLSFSettings.from_setting(settings)

    def bsub_args(self) -> List[str]:
        args = [self.settings.bsub_binary, "-K"]  # always use -K to block
        args.extend(["-o", self.settings.log_file if self.settings.log_file is not None else
            datetime.datetime.now().strftime("hammer-vlsi-bsub-%Y%m%d-%H%M%S.log")])  # always use -o to log to a file
        if self.settings.queue is not None:
            args.extend(["-q", self.settings.queue])
        if self.settings.num_cpus is not None:
            args.extend(["-n", "%d" % self.settings.num_cpus])
        args.extend(self.settings.extra_args)
        return args

    def submit(self, args: List[str], env: Dict[str, str],
               logger: HammerVLSILoggingContext, cwd: Optional[str] = None) -> Tuple[str, int]:
        # TODO fix output capturing

        prog_tag = self.get_program_tag(args)

        subprocess_format_str = 'Executing subprocess: {bsub_args} "{args}"'
        logger.debug(subprocess_format_str.format(bsub_args=' '.join(self.bsub_args()),
                                                  args=' '.join(args)))
        subprocess_logger = logger.context("Exec " + prog_tag)
        proc = subprocess.Popen(self.bsub_args() + [' '.join(args)],
                                shell=False, stderr=subprocess.STDOUT,
                                stdout=subprocess.PIPE, env=env, cwd=cwd)

        output_buf = ""
        # Log output and also capture output at the same time.
        so = proc.stdout
        assert so is not None
        while True:
            line = so.readline().decode("utf-8")
            if line != '':
                subprocess_logger.debug(line.rstrip())
                output_buf += line
            else:
                break
        # check errors
        proc.communicate()

        return output_buf, proc.returncode



# See LICENSE for licence details.

import os
import sys

# Try to use the system yaml if present, else use the HAMMER-shipped yaml.
try:
    import yaml
except ImportError:
    try:
        sys.path.append("src/tools/pyyaml/lib3")
        import yaml
    except ImportError:
        if "HAMMER_PYYAML_PATH" not in os.environ:
            print("pyyaml not found. Set $HAMMER_PYYAML_PATH to pyyaml/lib3", file=sys.stderr)
            sys.exit(1)
        else:
            sys.path.append(os.environ["HAMMER_PYYAML_PATH"])
            import yaml

import json

def convertArrays(o):
    """
    The YAML parser will take a list of substructures all named with an ints
    and create a python sub dict using those ints as keys. In JSON you can't
    have keys be anything other than strings even though that is valid in a
    python dict. This function recursively converts a python dict tree into
    one where any dicts where all the keys are ints are made into arrays. It
    makes sure the order is preserved.

    Adapted from https://raw.githubusercontent.com/vasil9v/yaml2json.py/c4cfa408bbab4f1a0cd3661d121237030f6bc0ca/yaml2json.py
    """
    if type(o) == list:
        for i in range(len(o)):
            o[i] = convertArrays(o[i])
        return o
    if type(o) == dict:
        allNums = True  # type: bool
        if len(o) == 0:
            # Don't convert anything if the dictionary is empty
            allNums = False
        else:
            # Check if it's the case that each key is an int
            for i in o:
                o[i] = convertArrays(o[i])
                if type(i) != int:
                    allNums = False
        # if so, convert it to an array
        if allNums:
            k = sorted(o.keys())
            arr = []
            for i in k:
                arr.append(o[i])
            return arr
    return o

def compare(o1, o2):
    """
    Recursively compares each item in a python dict tree to make sure they are
    the same.

    Adapted from https://raw.githubusercontent.com/vasil9v/yaml2json.py/c4cfa408bbab4f1a0cd3661d121237030f6bc0ca/yaml2json.py
    """
    if o1 == o2:
        return True # shortcut if they're the same primitive
    if type(o1) != type(o2):
        return False # shortcut if they're not the same type
    if type(o1) == type({}):
        k1 = sorted(o1.keys())
        k2 = sorted(o2.keys())
        if k1 == k2: # make sure the sorted keys of the 2 dicts match
            for i in o1: # recursively compare each sub item
                if not compare(o1[i], o2[i]):
                    return False
        else:
            return False
    if type(o1) == type([]): # compare each item in an array
        if len(o1) != len(o2):
            return False
        for i in range(len(o1)):
            if not compare(o1[i], o2[i]):
                return False
    return (o1 == o2)


def load_yaml(yamlStr: str) -> dict:
    """
    Load a YAML database as JSON.

    The input file is parsed as YAML and converted to a python dict tree, then
    that tree is converted to the JSON output. There is a check to make sure the
    two dict trees are structurally identical.

    :param yamlStr: A string containing the yaml database.
    :return: A dictionary object representing the yaml database.
    """
    obj = convertArrays(yaml.safe_load(yamlStr))
    # Note we are not using HammerJSONEncoder here to avoid a circular dependency, but this should never need have Decimals
    obj2 = json.loads(json.dumps(obj))
    if not compare(obj, obj2):
        raise ValueError("YAML -> JSON structures don't match: %s and %s do not match" % (str(obj), str(obj2)))
    else:
        if obj2 is None:
            # Loading a YAML file with nothing (except comments) can return None.
            return {}
        else:
            assert isinstance(obj2, dict), "Config databases should be a dictionary"
            return obj2



# common tool settings and functions for OpenROAD tools
#
# See LICENSE for licence details.

from decimal import Decimal
from functools import reduce
import os
import re
import shutil
from subprocess import Popen, PIPE
from textwrap import dedent as dd
from typing import List, Optional, Dict, Set, Union

from hammer.utils import add_dicts
from hammer.vlsi import HasSDCSupport, HammerSynthesisTool, \
    HammerPlaceAndRouteTool, HammerDRCTool, \
    PlacementConstraintType, TimeValue, \
    TCLTool, HammerTool
import hammer.tech as hammer_tech


class OpenROADTool(HasSDCSupport, TCLTool, HammerTool):
    """ Mix-in trait with functions useful for OpenROAD-flow tools."""

    @property
    def env_vars(self) -> Dict[str, str]:
        """
        Get the list of environment variables required for this tool.
        Note to subclasses: remember to include variables from 
        super().env_vars!
        """
        list_of_vars = self.get_setting("openroad.extra_env_vars")
        assert isinstance(list_of_vars, list)
        return reduce(add_dicts, [dict(super().env_vars)] + list_of_vars, {})

    def validate_openroad_installation(self) -> None:
        """
        make sure OPENROAD env-var is set, and klayout is in the path (since
        klayout is not installed with OPENROAD as of version 1.1.0. this 
        should be called in steps that actually run tools or touch filepaths
        """
        if not shutil.which("openroad"):
            raise Exception("openroad is not in PATH")
        if not shutil.which("klayout"):
            raise Exception("klayout is not in PATH")

    def openroad_flow_path(self) -> str:
        """return the root of the OpenROAD-flow installation"""
        self.validate_openroad_installation()
        return os.path.realpath(os.path.join(os.environ['OPENROAD'], "../.."))

    def openroad_flow_makefile_path(self) -> str:
        """drive OpenROAD-flow's top-level Makefile from hammer"""
        openroad = self.openroad_flow_path()
        return os.path.join(openroad, "flow/Makefile")

    def syn_rundir_path(self) -> str:
        # TODO: leaking info from driver into this tool. figure out a better
        # way to get "syn-rundir" than duplicating a constant
        return os.path.realpath(os.path.join(self.run_dir, "../syn-rundir"))

    def syn_objects_path(self) -> str:
        """where the intermediate synthesis output objects are located"""
        return os.path.join(self.syn_rundir_path(), "objects",
                            self.get_setting("vlsi.core.technology"),
                            self.top_module)

    def syn_results_path(self) -> str:
        """where the intermediate synthesis final results are located"""
        return os.path.join(self.syn_rundir_path(), "results",
                            self.get_setting("vlsi.core.technology"),
                            self.top_module)

    def design_config_path(self) -> str:
        """
        the initial design_config is written by the synthesis step. any other
        tool must refer to this design_config
        """
        return os.path.join(self.syn_rundir_path(), "design_config.mk")

    def par_rundir_path(self) -> str:
        # TODO: leaking info from driver into this tool. figure out a better
        # way to get "par-rundir" than duplicating a constant
        return os.path.realpath(os.path.join(self.run_dir, "../par-rundir"))

    def par_objects_path(self) -> str:
        """where the intermediate par output objects are located"""
        return os.path.join(self.par_rundir_path(), "objects",
                            self.get_setting("vlsi.core.technology"),
                            self.top_module)

    def par_results_path(self) -> str:
        """where the intermediate par final results are located"""
        return os.path.join(self.par_rundir_path(), "results",
                            self.get_setting("vlsi.core.technology"),
                            self.top_module)

    def version_number(self, version: str) -> int:
        """get OPENROAD-flow's version from one of its header files"""

        filepath = os.path.join(self.openroad_flow_path(),
                                "tools/OpenROAD/include/openroad/Version.hh")
        process = Popen(["grep", "OPENROAD_VERSION", filepath], stdout=PIPE)
        (output, err) = process.communicate()
        code = process.wait()
        assert code == 0, "grep failed with code {}".format(code)
        text = output.decode("utf-8")
        match = re.search(r"OPENROAD_VERSION \"(\d+)\.(\d+)\.(\d+)\"", text)
        if match is None:
            raise Exception("OPENROAD_VERSION could not be found!")
        return int(match.group(1)) * 10000 + \
               int(match.group(2)) * 100 + \
               int(match.group(3))

    def setup_openroad_rundir(self) -> bool:
        """
        OpenROAD expects several files/dirs in the current run_dir, so we 
        symlink them in from the OpenROAD-flow installation
        """
        # TODO: for now, just symlink in the read-only OpenROAD stuff, since
        # the $OPENROAD/flow/Makefile expects these in the current directory.
        # in the future, $OPENROAD/flow/Makefile should use ?= instead of =
        # for these symlinked dirs
        openroad = self.openroad_flow_path()
        for subpath in ["platforms", "scripts", "util", "test"]:
            src = os.path.join(openroad, "flow/{}".format(subpath))
            dst = os.path.join(self.run_dir, subpath)
            try:
                os.remove(dst)
            except:
                pass
            os.symlink(src, dst)
        return True


class OpenROADSynthesisTool(HammerSynthesisTool, OpenROADTool):
    """ Mix-in trait with functions for OpenROAD-flow synthesis tools."""

    # =========================================================================
    # overrides from parent classes
    # =========================================================================
    # @property
    # don't overrride this!
    # def post_synth_sdc(self) -> Optional[str]:
    #     return None

    # =========================================================================
    # OpenROAD synthesis-specific stuff
    # =========================================================================
    def _clock_period_value(self) -> str:
        """this string is used in the makefile fragment used by OpenROAD"""

        assert len(self.get_clock_ports()) == 1, "openroad only supports 1 root clock"
        return str(self.get_clock_ports()[0].period.value_in_units("ns"))

    def _floorplan_bbox(self) -> str:
        """this string is used in the makefile fragment used by OpenROAD"""

        floorplan_constraints = self.get_placement_constraints()
        for constraint in floorplan_constraints:
            new_path = "/".join(constraint.path.split("/")[1:])

            if new_path == "":
                assert constraint.type == PlacementConstraintType.TopLevel, \
                    "Top must be a top-level/chip size constraint"
                margins = constraint.margins
                assert margins is not None
                # Set top-level chip dimensions.
                return "{} {} {} {}".format(
                    constraint.x, constraint.y,
                    constraint.width, constraint.height)

        raise Exception("no top-level placement constraint was found")

    def create_design_config(self) -> bool:
        """
        the design-config is the main configuration for the OpenROAD built-in
        scripts. initially, we are using OpenROAD's tool scripts as-is, so
        we need to create a design-config that their scripts understand. the
        synthesis tool is the only tool that will write this
        """
        design_config = self.design_config_path()

        # Load input files and check that they are all Verilog.
        if not self.check_input_files([".v", ".sv"]):
            return False
        abspath_input_files = list(map(lambda name:
                                       os.path.join(os.getcwd(), name), self.input_files))

        # Add any verilog_synth wrappers (which are needed in some 
        # technologies e.g. for SRAMs) which need to be synthesized.
        abspath_input_files += self.technology.read_libs([
            hammer_tech.filters.verilog_synth_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)

        # Generate constraints
        input_sdc = os.path.join(self.run_dir, "input.sdc")
        unit = self.get_time_unit().value_prefix + self.get_time_unit().unit
        with open(input_sdc, "w") as f:
            f.write("set_units -time {}\n".format(unit))
            f.write(self.sdc_clock_constraints)
            f.write("\n")
            f.write(self.sdc_pin_constraints)

        # TODO: i am blindly reading in all libs for all corners. but this is
        #       not a performance issue for nangate45
        extra_lefs = set([extra_lib.library.lef_file for extra_lib in self.technology.get_extra_libraries()
                          if extra_lib.library.lef_file is not None])
        extra_libs = set([extra_lib.library.nldm_liberty_file for extra_lib in self.technology.get_extra_libraries()
                          if extra_lib.library.nldm_liberty_file is not None])

        with open(design_config, "w") as f:
            f.write(dd("""
            export DESIGN_NICKNAME = {design}
            export DESIGN_NAME     = {design}
            export PLATFORM        = {node}
            export VERILOG_FILES   = {verilogs}
            export SDC_FILE        = {sdc}

            export ADDITIONAL_LEFS = {extra_lefs}
            export ADDITIONAL_LIBS = {extra_libs}

            # These values must be multiples of placement site, which is
            # (x=0.19 y=1.4) for nangate45
            export DIE_AREA    = {die_area}
            export CORE_AREA   = {core_area}

            export CLOCK_PERIOD = {period}

            """.format(
                design=self.top_module,
                node=self.get_setting("vlsi.core.technology"),
                verilogs=" ".join(abspath_input_files),
                sdc=input_sdc,
                extra_lefs=" ".join(extra_lefs),
                extra_libs=" ".join(extra_libs),
                die_area=self._floorplan_bbox(),
                core_area=self._floorplan_bbox(),
                period=self._clock_period_value(),
            )))
        return True


class OpenROADPlaceAndRouteTool(HammerPlaceAndRouteTool, OpenROADTool):
    """ Mix-in trait with functions for OpenROAD-flow par tools."""

    # =========================================================================
    # overrides from parent classes
    # =========================================================================
    def specify_power_straps(self, layer_name: str,
                             bottom_via_layer_name: str, blockage_spacing: Decimal, pitch: Decimal,
                             width: Decimal, spacing: Decimal, offset: Decimal,
                             bbox: Optional[List[Decimal]], nets: List[str],
                             add_pins: bool) -> List[str]:
        # TODO: currently using openroad's powerstrap script
        return []

    def specify_std_cell_power_straps(self, blockage_spacing: Decimal,
                                      bbox: Optional[List[Decimal]], nets: List[str]) -> List[str]:
        # TODO: currently using openroad's powerstrap script
        return []


class OpenROADDRCTool(HammerDRCTool, OpenROADTool):
    """ Mix-in trait with functions for OpenROAD-flow drc tools."""

    # =========================================================================
    # overrides from parent classes
    # =========================================================================
    @property
    def post_synth_sdc(self) -> Optional[str]:
        return None

    def globally_waived_drc_rules(self) -> List[str]:
        # TODO: i suppose the actual implementation should implement this
        return []

    def drc_results_pre_waived(self) -> Dict[str, int]:
        # TODO: i suppose the actual implementation should implement this
        return {}



#!/usr/bin/env python3
# type: ignore
#   tell mypy to ignore this file during typechecking
# -*- coding: utf-8 -*-
#
#  Generate Hammer Sky130 tech plugin file: sky130.tech.json
#
#  See LICENSE for licence details.
import sys
import json
import os

use_nda_files=True
library='sky130_fd_sc_hd'

def main(args) -> int:
    if len(args) != 3:
        print("Usage: ./sky130-tech-gen.py /path/to/sky130A sky130.tech.json")
        return 1

    SKY130A = sys.argv[1]

    if use_nda_files:
        with open('sky130-tech-gen-files/beginning_nda.json', 'r') as f: data = json.load(f)
    else:
        with open('sky130-tech-gen-files/beginning.json', 'r') as f: data = json.load(f)

    with open('sky130-tech-gen-files/cells.json', 'r') as f:
        cells = json.load(f)
    data["physical_only_cells_list"] = cells["physical_only_cells_list"]
    data["dont_use_list"] = cells["dont_use_list"]
    data["special_cells"] = cells["special_cells"]

    SKYWATER_LIBS = os.path.join('$SKY130A', 'libs.ref', library)
    LIBRARY_PATH  = os.path.join(  SKY130A,  'libs.ref', library, 'lib')
    lib_corner_files=os.listdir(LIBRARY_PATH)
    for cornerfilename in lib_corner_files:
        if (not (library in cornerfilename) ) : continue
        if ('ccsnoise' in cornerfilename): continue # ignore duplicate corner.lib/corner_ccsnoise.lib files

        tmp = cornerfilename.replace('.lib','')
        if (tmp+'_ccsnoise.lib' in lib_corner_files): 
            cornerfilename=tmp+'_ccsnoise.lib' # use ccsnoise version of lib file

        cornername = tmp.split('__')[1]
        cornerparts = cornername.split('_')

        speed = cornerparts[0]
        if (speed == 'ff'): speed = 'fast'
        if (speed == 'tt'): speed = 'typical'
        if (speed == 'ss'): speed = 'slow'

        temp = cornerparts[1]
        temp = temp.replace('n','-')
        temp = temp.split('C')[0]+' C'

        vdd = cornerparts[2]
        vdd = vdd.split('v')[0]+'.'+vdd.split('v')[1]+' V'

        lib_entry = {
            "nldm_liberty_file":  os.path.join(SKYWATER_LIBS,'lib', cornerfilename),
            "verilog_sim":        os.path.join('cache',             library+'.v'),
            "lef_file":           os.path.join(SKYWATER_LIBS,'lef', library+'.lef'),
            "spice_file":         os.path.join('cache',             library+'.cdl'),
            "gds_file":           os.path.join(SKYWATER_LIBS,'gds', library+'.gds'),
            "corner": {
                "nmos": speed,
                "pmos": speed,
                "temperature": temp
            },
            "supplies": {
                "VDD": vdd,
                "GND": "0 V"
            },
            "provides": [
                {
                "lib_type": "stdcell",
                "vt": "RVT" 
                }
            ]
        }

        data["libraries"].append(lib_entry)

    with open('sky130-tech-gen-files/stackups.json', 'r') as f:
        stackups = json.load(f)
    data["stackups"] = [stackups]

    with open('sky130-tech-gen-files/sites.json', 'r') as f:
        sites = json.load(f)
    data["sites"] = sites["sites"]

    with open(sys.argv[2], 'w') as f:
        json.dump(data, f, indent=2)
    
    return 0

if __name__ == '__main__':
  sys.exit(main(sys.argv))



# Magic DRC plugin for Hammer
#
# See LICENSE for licence details.

import os
from textwrap import dedent as dd
from typing import List, Optional, Dict, Any

from hammer.logging import HammerVLSILogging
from hammer.utils import deepdict
from hammer.vlsi import HammerToolStep
from hammer.vlsi import HammerDRCTool, TCLTool

class Magic(HammerDRCTool, TCLTool):

    #=========================================================================
    # overrides from parent classes
    #=========================================================================
    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.init_design,
            self.run_drc
        ])

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_magic()

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        return new_dict

    def fill_outputs(self) -> bool:
        return True

    def tool_config_prefix(self) -> str:
        return "drc.magic"

    def drc_results_pre_waived(self) -> Dict[str, int]:
        return {}

    def globally_waived_drc_rules(self) -> List[str]:
        return []

    def version_number(self, version:str) -> int:
        """Get version from magic bin"""
        version = self.run_executable([self.get_setting("drc.magic.magic_bin"), "--version"])
        return int(version.replace(".", ""))

    @property
    def generated_scripts_dir(self) -> str:
        return os.path.join(self.run_dir, "generated-scripts")

    @property
    def view_drc_script(self) -> str:
        return os.path.join(self.generated_scripts_dir, "view_drc")

    @property
    def view_drc_tcl(self) -> str:
        return os.path.join(self.generated_scripts_dir, "view_drc.tcl")

    #=========================================================================
    # useful subroutines
    #=========================================================================

    def run_magic(self) -> bool:
        self.append("quit")

        run_script = os.path.join(self.run_dir, "drc.tcl")
        with open(run_script, "w") as f:
            f.write("\n".join(self.output))

        args = [self.get_setting("drc.magic.magic_bin")]
        rcfile = self.get_setting("drc.magic.rcfile")
        techfile = self.get_drc_decks()

        if rcfile is not None:
            args.extend(["-rcfile", rcfile])
        else:
            # DRC deck should be the tech file. There should only be 1.
            if len(techfile) == 0 or len(techfile) > 1:
                self.logger.error("None or more than 1 tech file (DRC deck) found. netgen only supports 1.")
            args.extend(["-T", techfile[0]])

        """
        Create view_drc script. This opens interactive window but has to run DRC
        all over again because there is no DRC database that can be loaded in.
        """
        os.makedirs(self.generated_scripts_dir, exist_ok=True)
        with open(self.view_drc_script, "w") as f:
            f.write(dd("""
            cd {run_dir}
            {args} {gds} {run_script}
            """.format(
                run_dir = self.run_dir,
                args = " ".join(args),
                gds = self.layout_file,
                run_script = self.view_drc_tcl
                )).strip())
        os.chmod(self.view_drc_script, 0o755)
        with open(self.view_drc_tcl, "w") as f:
            f.write(dd("""
            gds read {gds}
            load {top}
            select top cell
            drc check
            """.format(gds = self.layout_file, top = self.top_module)).strip())

        # Finally append the no GUI options and full Tcl run script
        args.extend(["-noconsole", "-dnull", run_script])

        if bool(self.get_setting("drc.magic.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + \
                             " ".join(args))
        else:
            # Temporarily disable colors/tag to make run output more readable
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable(args, cwd=self.run_dir) # TODO: check for errors
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

    #========================================================================
    # drc main steps
    #========================================================================
    def init_design(self) -> bool:
        """Read design and set up results outputs"""
        self.append("gds read " + self.layout_file)
        self.append("load " + self.top_module)
        self.append("select top cell")
        self.append("set oscale [cif scale out]")

        # Results output is taken from output of certain commands
        # Hooks that also write to it should use the $fout variable
        self.append("set fout [open drc.out w]")

        self.append(self.get_additional_drc_text())
        return True

    def run_drc(self) -> bool:
        """Run DRC and dump error boxes"""
        self.append('puts "Running DRC..."')
        self.append("drc check")
        self.append("set flat_count 0")

        # Adapted from OpenLane
        self.append('puts $fout "----------------------------------------"')
        self.append("set drcresult [drc listall why]")
        self.append('''
        foreach {errtype coordlist} $drcresult {
            puts $fout $errtype
            puts $fout "----------------------------------------"
            foreach coord $coordlist {
                set bllx [expr {$oscale * [lindex $coord 0]}]
                set blly [expr {$oscale * [lindex $coord 1]}]
                set burx [expr {$oscale * [lindex $coord 2]}]
                set bury [expr {$oscale * [lindex $coord 3]}]
                set coords [format " %.3fum %.3fum %.3fum %.3fum" $bllx $blly $burx $bury]
                puts $fout "$coords"
                incr flat_count
            }
            puts $fout "----------------------------------------"
        }
        puts $fout ""
        close $fout
        ''')

        # Error counts
        # TODO: hierarchical counts don't work in noconsole mode...
        #self.append('puts "Hierarchical DRC error counts:"')
        #self.append("puts [drc listall count]")
        self.append('puts "Flat DRC error count: $flat_count"')
        return True

tool = Magic



# OpenROAD-flow's nangate45 plugin for Hammer.
#
# See LICENSE for licence details.

from hammer.tech import HammerTechnology

class Nangate45Tech(HammerTechnology):
    """
    Override the HammerTechnology used in `hammer_tech.py`
    This class is loaded by function `load_from_json`, and will pass 
    the `try` in `importlib`.
    """

tech = Nangate45Tech()



# OpenROAD-flow klayout drc plugin for Hammer
#
# See LICENSE for licence details.

import glob
import os
from textwrap import dedent as dd
from typing import List, Optional, Dict, Any

from hammer.logging import HammerVLSILogging
from hammer.utils import deepdict
from hammer.vlsi import HammerToolStep
from hammer.vlsi.vendor import OpenROADDRCTool

class OpenROADDRC(OpenROADDRCTool):

    #=========================================================================
    # overrides from parent classes
    #=========================================================================
    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.init_design,
            self.run_drc_tool,
        ])

    def do_pre_steps(self, first_step: HammerToolStep) -> bool:
        assert super().do_pre_steps(first_step)
        self.cmds = []  # type: List[str]
        return True

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_openroad()

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        new_dict.update({})  # TODO: stuffs
        return new_dict

    def fill_outputs(self) -> bool:
        return True

    def tool_config_prefix(self) -> str:
        return "drc.openroad"

    #=========================================================================
    # useful subroutines
    #=========================================================================

    def run_openroad(self) -> bool:
        run_script = os.path.join(self.run_dir, "drc.sh")

        self.validate_openroad_installation()
        self.setup_openroad_rundir()

        with open(run_script, "w") as f:
            f.write(dd("""\
              #!/bin/bash
              cd "{rundir}"
              mkdir -p logs/{tech}/{name}
              mkdir -p objects/{tech}/{name}
              mkdir -p reports/{tech}/{name}
              mkdir -p results/{tech}/{name}
            """.format(
              rundir=self.run_dir,
              tech=self.get_setting("vlsi.core.technology"),
              name=self.top_module,
            )))
            f.write("\n".join(self.cmds))
        os.chmod(run_script, 0o755)

        if bool(self.get_setting("drc.openroad.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + \
                             " ".join(self.cmds))
        else:
            # Temporarily disable colors/tag to make run output more readable
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable([run_script]) # TODO: check for errors
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

    #========================================================================
    # drc main steps
    #========================================================================
    def init_design(self) -> bool:
        # TODO: currently hardlinking the par outputs, otherwise the
        # OpenROAD-flow's default makefile will rebuild syn in this directory
        par_dirs = {
          "results": self.par_results_path(),
          "objects": self.par_objects_path()
        }
        for par_dir in ["results", "objects"]:
          for src in glob.glob(os.path.join(par_dirs[par_dir], "*")):
              dst = "{par_dir}/{tech}/{name}/{base}".format(
                  par_dir=par_dir,
                  tech=self.get_setting("vlsi.core.technology"),
                  name=self.top_module,
                  base=os.path.basename(src))
              self.cmds += [
                  "rm -f {}".format(dst),
                  "ln {} {}".format(src, dst)
              ]
        return True

    def run_drc_tool(self) -> bool:
        # TODO: currently using OpenROAD's default drc script
        self.cmds += [
          "make DESIGN_CONFIG={conf} -f {make} drc".format(
            conf=self.design_config_path(),
            make=self.openroad_flow_makefile_path()
        )]
        return True

tool = OpenROADDRC



#  Mock hammer-vlsi place-and-route plugin to help test Hammer infrastructure
#  without proprietary/NDAed tools.
#  NOT FOR EXTERNAL/PUBLIC USE.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerPlaceAndRouteTool, DummyHammerTool, HammerToolStep, deepdict
from hammer.config import HammerJSONEncoder

from typing import Dict, List, Any, Optional
from decimal import Decimal

import os
import json


class MockPlaceAndRoute(HammerPlaceAndRouteTool, DummyHammerTool):

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        new_dict.update({})  # TODO: stuffs
        return new_dict

    def temp_file(self, filename: str) -> str:
        """Helper function to get the full path to a filename under temp_folder."""
        if self.get_setting("par.mockpar.temp_folder", nullvalue="") == "":
            raise ValueError("par.mockpar.temp_folder is not set correctly")
        return os.path.join(self.get_setting("par.mockpar.temp_folder"), filename)

    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.power_straps
        ])

    def power_straps(self) -> bool:
        power_straps_tcl = os.path.join(self.run_dir, "power_straps.tcl")
        with open(power_straps_tcl, "w") as f:
            f.write("\n".join(self.create_power_straps_tcl()))
        return True

    def parse_mock_power_straps_file(self) -> List[Dict[str, Any]]:
        power_straps_tcl = os.path.join(self.run_dir, "power_straps.tcl")
        output = []  # type: List[Dict[str, Any]]
        with open(power_straps_tcl, "r") as f:
            for line in f.readlines():
                output.append(json.loads(line))
        return output

    def specify_power_straps(self, layer_name: str, bottom_via_layer_name: str, blockage_spacing: Decimal, pitch: Decimal, width: Decimal, spacing: Decimal, offset: Decimal, bbox: Optional[List[Decimal]], nets: List[str], add_pins: bool) -> List[str]:
        self._power_straps_check_index(layer_name)
        output_dict = {
            "layer_name": layer_name,
            "bottom_via_layer_name": bottom_via_layer_name,
            "blockage_spacing": str(blockage_spacing),
            "pitch": str(pitch),
            "width": str(width),
            "spacing": str(spacing),
            "offset": str(offset),
            "bbox": [] if bbox is None else list(map(str, bbox)),
            "nets": list(map(str, nets)),
            "add_pins": add_pins
        }
        return [json.dumps(output_dict, cls=HammerJSONEncoder)]

    def specify_std_cell_power_straps(self, blockage_spacing: Decimal, bbox: Optional[List[Decimal]], nets: List[str]) -> List[str]:
        layer_name = self.get_setting("technology.core.std_cell_rail_layer")
        self._power_straps_check_index(layer_name)
        output_dict = {
            "layer_name": layer_name,
            "tap_cell_name": self.get_setting("technology.core.tap_cell_rail_reference"),
            "bbox": [] if bbox is None else list(map(str, bbox)),
            "nets": list(map(str, nets))
        }
        return [json.dumps(output_dict, cls=HammerJSONEncoder)]

    def fill_outputs(self) -> bool:
        self.output_gds = "/dev/null"
        self.output_ilms = []
        self.output_netlist = "/dev/null"
        self.output_sim_netlist = "/dev/null"
        self.hcells_list = []
        return True


tool = MockPlaceAndRoute



#!/usr/bin/env python3
#  generate_properties.py
#
#  Helper script to generate properties for hammer-vlsi tool classes.
#
#  See LICENSE for licence details.

import argparse
import os
import re
import sys
from collections import namedtuple
from typing import Dict

InterfaceVar = namedtuple("InterfaceVar", 'name type desc')

Interface = namedtuple("Interface", 'module filename inputs outputs')


def isinstance_check(t: str) -> str:
    return "isinstance(value, {t})".format(t=t)


def generate_from_list(template: str, lst) -> list:
    def format_var(var):
        attr_error_logic = """raise ValueError("Nothing set for the {var_desc} yet")"""

        if var.type.startswith("Iterable"):
            var_type_instance_check = isinstance_check("Iterable")
        elif var.type.startswith("List"):
            var_type_instance_check = isinstance_check("List")
        elif var.type.startswith("Optional"):
            m = re.search(r"Optional\[(\S+)\]", var.type)
            subtype = str(m.group(1))
            var_type_instance_check = isinstance_check(subtype) + " or (value is None)"
            attr_error_logic = "return None"
        else:
            var_type_instance_check = isinstance_check(var.type)

        t = template.replace("[[attr_error_logic]]", attr_error_logic)

        return t.format(var_name=var.name, var_type=var.type, var_desc=var.desc,
                        var_type_instance_check=var_type_instance_check)

    return list(map(format_var, lst))


# Cache of files being modified.
file_cache = {}  # type: Dict[str, str]


def get_full_filename(filename: str) -> str:
    return os.path.join(os.path.dirname(__file__), filename)


def generate_interface(interface: Interface) -> None:
    template = """
    @property
    def {var_name}(self) -> {var_type}:
        \"""
        Get the {var_desc}.

        :return: The {var_desc}.
        \"""
        try:
            return self.attr_getter("_{var_name}", None)
        except AttributeError:
            [[attr_error_logic]]

    @{var_name}.setter
    def {var_name}(self, value: {var_type}) -> None:
        \"""Set the {var_desc}.\"""
        if not ({var_type_instance_check}):
            raise TypeError("{var_name} must be a {var_type}")
        self.attr_setter("_{var_name}", value)
"""
    start_key = "    ### Generated interface %s ###" % (interface.module)
    end_key = "    ### END Generated interface %s ###" % (interface.module)

    output = []
    output.append(start_key)
    output.append("    ### DO NOT MODIFY THIS CODE, EDIT %s INSTEAD ###" % (os.path.basename(__file__)))
    output.append("    ### Inputs ###")
    output.extend(generate_from_list(template, interface.inputs))
    output.append("")
    output.append("    ### Outputs ###")
    output.extend(generate_from_list(template, interface.outputs))
    output.append(end_key)

    filename = get_full_filename(interface.filename)

    if filename not in file_cache:
        with open(filename, "r") as f:
            file_cache[filename] = str(f.read())
    contents = file_cache[filename]

    new_contents = re.sub(re.escape(start_key) + ".*" + re.escape(end_key), "\n".join(output), contents,
                          flags=re.MULTILINE | re.DOTALL)

    file_cache[filename] = new_contents


def main(args) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', "--dry_run", action="store_true", required=False)
    parser.add_argument("-f", "--file", required=False, default="",
                        help='File to write/print. If unset, will write/print all files.')
    parsed_args = parser.parse_args(args[1:])

    HammerSynthesisTool = Interface(module="HammerSynthesisTool",
                                    filename="vlsi/hammer_vlsi_impl.py",
                                    inputs=[
                                        InterfaceVar("input_files", "List[str]",
                                                     "input collection of source RTL files (e.g. *.v)")
                                    ],
                                    outputs=[
                                        InterfaceVar("output_files", "List[str]",
                                                     "output collection of mapped (post-synthesis) RTL files"),
                                        InterfaceVar("output_sdc", "str",
                                                     "(optional) output post-synthesis SDC constraints file"),
                                        InterfaceVar("output_all_regs", "str",
                                                     "path to output list of all registers in the design with output pin for gate level simulation"),
                                        InterfaceVar("output_seq_cells", "str",
                                                     "path to output collection of all sequential standard cells in design"),
                                        InterfaceVar("sdf_file", "str",
                                                     "output SDF file to be read for timing annotated gate level sims")
                                        # TODO: model CAD junk
                                    ]
                                    )

    HammerPlaceAndRouteTool = Interface(module="HammerPlaceAndRouteTool",
                                        filename="vlsi/hammer_vlsi_impl.py",
                                        inputs=[
                                            InterfaceVar("input_files", "List[str]",
                                                         "input post-synthesis netlist files"),
                                            InterfaceVar("post_synth_sdc", "Optional[str]",
                                                         "(optional) input post-synthesis SDC constraint file"),
                                        ],
                                        outputs=[
                                            # TODO: model more CAD junk

                                            # e.g. par-rundir/TopModuleILMDir/mmmc/ilm_data/TopModule. Has a bunch of files TopModule_postRoute*
                                            InterfaceVar("output_ilms", "List[ILMStruct]",
                                                         "(optional) output ILM information for hierarchical mode"),
                                            InterfaceVar("output_gds", "str", "path to the output GDS file"),
                                            InterfaceVar("output_netlist", "str", "path to the output netlist file"),
                                            InterfaceVar("output_sim_netlist", "str", "path to the output simulation netlist file"),
                                            InterfaceVar("hcells_list", "List[str]",
                                                         "list of cells to explicitly map hierarchically in LVS"),
                                            InterfaceVar("output_all_regs", "str",
                                                         "path to output list of all registers in the design with output pin for gate level simulation"),
                                            InterfaceVar("output_seq_cells", "str",
                                                         "path to output collection of all sequential standard cells in design"),
                                            InterfaceVar("sdf_file", "str",
                                                     "output SDF file to be read for timing annotated gate level sims")
                                            # TODO: add individual parts of the ILM (e.g. verilog, sdc, spef, etc) for cross-tool compatibility?
                                        ]
                                        )

    HammerSRAMGeneratorTool = Interface(module="HammerSRAMGeneratorTool",
                                       filename="vlsi/hammer_vlsi_impl.py",
                                       inputs=[
                                           InterfaceVar("input_parameters", "List[SRAMParameters]",
                                                        "input sram parameters to be generated")
                                           ],
                                       outputs=[
                                           InterfaceVar("output_libraries", "List[ExtraLibrary]",
                                                        "list of the hammer tech libraries corresponding to generated srams")
                                           ]
                                       )

    HammerDRCTool = Interface(module="HammerDRCTool",
                              filename="vlsi/hammer_vlsi_impl.py",
                              inputs=[
                                  InterfaceVar("layout_file", "str", "path to the input layout file (e.g. a *.gds)")
                              ],
                              outputs=[]
                              )

    HammerLVSTool = Interface(module="HammerLVSTool",
                              filename="vlsi/hammer_vlsi_impl.py",
                              inputs=[
                                  InterfaceVar("layout_file", "str", "path to the input layout file (e.g. a *.gds)"),
                                  InterfaceVar("schematic_files", "List[str]",
                                               "path to the input SPICE or Verilog schematic files (e.g. *.v or *.spi)"),
                                  InterfaceVar("hcells_list", "List[str]",
                                               "list of cells to explicitly map hierarchically in LVS"),
                                  InterfaceVar("ilms", "List[ILMStruct]",
                                               "list of (optional) input ILM information for hierarchical mode")
                              ],
                              outputs=[]
                              )

    HammerSimTool = Interface(module="HammerSimTool",
                              filename="vlsi/hammer_vlsi_impl.py",
                              inputs=[
                                  InterfaceVar("top_module", "str", "top RTL module"),
                                  InterfaceVar("input_files", "List[str]", "paths to input verilog files"),
                                  InterfaceVar("all_regs", "str", "path to list of all registers in the design with output pin"),
                                  InterfaceVar("seq_cells", "str", "path to collection of all sequential standard cells in design"),
                                  InterfaceVar("sdf_file", "Optional[str]", "optional SDF file needed for timing annotated gate level sims")
                              ],
                              outputs=[
                                  InterfaceVar("output_waveforms", "List[str]", "paths to output waveforms"),
                                  InterfaceVar("output_saifs", "List[str]", "paths to output activity files"),
                                  InterfaceVar("output_top_module", "str", "top RTL module"),
                                  InterfaceVar("output_tb_name", "str", "sim testbench name"),
                                  InterfaceVar("output_tb_dut", "str", "sim DUT instance name"),
                                  InterfaceVar("output_level", "str", "simulation flow level")
                              ]
                              )
    HammerPowerTool = Interface(module="HammerPowerTool",
                                filename="vlsi/hammer_vlsi_impl.py",
                                inputs=[
                                  InterfaceVar("flow_database", "str", "path to syn or par database for power analysis"),
                                  InterfaceVar("input_files", "List[str]", "paths to RTL input files or design netlist"),
                                  InterfaceVar("spefs", "List[str]", "list of spef files for power anlaysis"),
                                  InterfaceVar("sdc", "Optional[str]","(optional) input SDC constraint file"),
                                  InterfaceVar("waveforms", "List[str]", "list of waveform dump files for dynamic power analysis"),
                                  InterfaceVar("saifs", "List[str]", "list of activity files for dynamic power analysis"),
                                  InterfaceVar("top_module", "str", "top RTL module"),
                                  InterfaceVar("tb_name", "str", "testbench name"),
                                  InterfaceVar("tb_dut", "str", "DUT instance name")
                                ],
                                outputs=[]
                                )
    HammerFormalTool = Interface(module="HammerFormalTool",
                            filename="vlsi/hammer_vlsi_impl.py",
                            inputs=[
                                InterfaceVar("check", "str",
                                    "formal verification check type to run"),
                                InterfaceVar("input_files", "List[str]",
                                    "input collection of implementation design files"),
                                InterfaceVar("reference_files", "List[str]",
                                    "input collection of reference design files"),
                                InterfaceVar("top_module", "str", "top RTL module"),
                                InterfaceVar("post_synth_sdc", "Optional[str]",
                                    "(optional) input post-synthesis SDC constraint file")
                            ],
                            outputs=[]
                            )
    HammerTimingTool = Interface(module="HammerTimingTool",
                            filename="vlsi/hammer_vlsi_impl.py",
                            inputs=[
                                InterfaceVar("input_files", "List[str]",
                                    "input collection of design files"),
                                InterfaceVar("top_module", "str", "top RTL module"),
                                InterfaceVar("post_synth_sdc", "Optional[str]",
                                    "(optional) input post-synthesis SDC constraint file"),
                                InterfaceVar("spefs", "Optional[List]",
                                    "(optional) list of SPEF files"),
                                InterfaceVar("sdf_file", "Optional[str]",
                                    "(optional) input SDF file")
                            ],
                            outputs=[]
                            )
    HammerPCBDeliverableTool = Interface(module="HammerPCBDeliverableTool",
                                       filename="vlsi/hammer_vlsi_impl.py",
                                       inputs=[],
                                       outputs=[
                                           InterfaceVar("output_footprints", "List[str]",
                                                        "list of the PCB footprint files for the project"),
                                           InterfaceVar("output_schematic_symbols", "List[str]",
                                                        "list of the PCB schematic symbol files for the project"),
                                           ]
                                       )

    dry_run = parsed_args.dry_run
    selected_file = str(parsed_args.file)

    generate_interface(HammerSynthesisTool)
    generate_interface(HammerPlaceAndRouteTool)
    generate_interface(HammerDRCTool)
    generate_interface(HammerLVSTool)
    generate_interface(HammerSRAMGeneratorTool)
    generate_interface(HammerSimTool)
    generate_interface(HammerPowerTool)
    generate_interface(HammerFormalTool)
    generate_interface(HammerTimingTool)
    generate_interface(HammerPCBDeliverableTool)

    if selected_file == "":
        # Export all files
        # print without extra newline
        for filename, contents in file_cache.items():
            if dry_run:
                print(contents, end='')
            else:
                with open(filename, "w") as f:
                    f.write(contents)
    else:
        # Export selected file
        contents = file_cache[get_full_filename(selected_file)]
        if dry_run:
            print(contents, end='')
        else:
            with open(filename, "w") as f:  # type: ignore
                f.write(contents)

    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv))
